from sys import getsizeof
from ipaddress import (
    ip_address,
    summarize_address_range,
    IPv4Network,
    IPv6Network
)
from typing import Optional
from typing_extensions import TypedDict
from dateutil.parser import parse as dateutil_parse
from .errors import BootstrapError, ParseError
from .logger import get_logger

log = get_logger('parser')


def clean(s):
    if s is None:
        s = ''
    if not isinstance(s, str):
        s = str(s)
    return s.strip()


class VCardArrayDataDict(TypedDict, total=False):
    name: str
    email: str
    tel: str


class Parser:
    """
        A Parser extracts the most useful information from an RDAP response for each
        response type (ip, domain etc.) and returns it as a flat dictionary. This
        parent class extracts generic information available to all entity types where
        available.
    """

    def __init__(self, bootstrap, raw_data, query, using_overrides=False):
        self.bootstrap = bootstrap
        self.raw_data = raw_data
        self.query = str(query)
        self.parsed = {}
        self.using_overrides = bool(using_overrides)
        self.extract_handle()
        # As a basic check every object must have at least a handle set
        if not self.parsed['handle']:
            # Permit overridden endpoints to not return a handle
            if not self.using_overrides:
                raise ParseError(
                    f'Failed to parse any meaningful data to find a '
                    f'handle in raw data: {self.raw_data}'
                )
        self.extract_parent_handle()
        self.extract_name()
        self.extract_whois_server()
        self.extract_response_type()
        self.extract_notices()
        self.extract_description()
        self.extract_dates()
        self.extract_self_link()
        self.extract_entities()

    def parse(self):
        raise NotImplementedError('parse must be implemented')

    def parse_vcard_array(self, vcard) -> Optional[VCardArrayDataDict]:
        '''
            Extract useful summary information from a VCard array.
        '''
        if not isinstance(vcard, list):
            return None
        if len(vcard) != 2:
            return None
        card_field, card_data = vcard
        if card_field != 'vcard':
            return None
        v_card_array_data_dict = VCardArrayDataDict()

        for field in card_data:
            if len(field) != 4:
                continue
            entry_field, entry_data, entry_type, entry_label = field

            if entry_type != 'text':
                continue
            elif entry_field == 'fn':
                v_card_array_data_dict["name"] = clean(entry_label)
            elif entry_field == 'org':
                v_card_array_data_dict["name"] = clean(entry_label)
            elif entry_field == 'email':
                v_card_array_data_dict["email"] = clean(entry_label)
            elif entry_field == 'tel':
                v_card_array_data_dict["tel"] = clean(entry_label)

        return v_card_array_data_dict or None

    def extract_handle(self):
        self.parsed['handle'] = clean(self.raw_data.get('handle', '')).upper()

    def extract_parent_handle(self):
        self.parsed['parent_handle'] = clean(
            self.raw_data.get('parentHandle', '')).upper()

    def extract_name(self):
        self.parsed['name'] = clean(self.raw_data.get('name', ''))

    def extract_whois_server(self):
        self.parsed['whois_server'] = clean(self.raw_data.get('port43', ''))

    def extract_response_type(self):
        self.parsed['type'] = clean(self.raw_data.get('objectClassName', ''))

    def extract_notices(self):
        self.parsed['terms_of_service_url'] = ''
        self.parsed['copyright_notice'] = ''
        for notice in self.raw_data.get('notices', []):
            title = clean(notice.get('title', '')).lower()
            if title in (
                    'terms of service', 'terms of use', 'terms and conditions'
            ):
                links = notice.get('links', [])
                try:
                    link = links[0]
                except (IndexError, KeyError):
                    continue
                self.parsed['terms_of_service_url'] = clean(
                    link.get('href', '')).strip()
            elif title == 'copyright notice':
                descriptions = notice.get('description', [])
                try:
                    description = descriptions[0]
                except IndexError:
                    continue
                self.parsed['copyright_notice'] = clean(description)

    def extract_description(self):
        self.parsed['description'] = []
        remarks = self.raw_data.get('remarks', [])
        for remark in remarks:
            title = clean(remark.get('title', '')).lower()
            description = remark.get('description', [])
            if not description:
                continue
            if len(remarks) == 1:
                # There is only one remark, use it
                self.parsed['description'] = description
            elif title == 'description':
                # Multiple remarks, add only the description
                self.parsed['description'] = description

    def extract_dates(self):
        self.parsed['last_changed_date'] = None
        self.parsed['registration_date'] = None
        self.parsed['expiration_date'] = None
        for event in self.raw_data.get('events', []):
            action = event.get('eventAction').strip().lower()
            if action == 'last changed':
                last_changed_date = event.get('eventDate', '')
                if last_changed_date:
                    self.parsed['last_changed_date'] = dateutil_parse(last_changed_date)
            elif action == 'registration':
                registration_date = event.get('eventDate', '')
                if registration_date:
                    self.parsed['registration_date'] = dateutil_parse(registration_date)
            elif action == 'expiration':
                expiration_date = event.get('eventDate', '')
                if expiration_date:
                    self.parsed['expiration_date'] = dateutil_parse(expiration_date)

    def extract_self_link(self):
        self.parsed['url'] = ''
        for link in self.raw_data.get('links', []):
            if link.get('rel', '').strip().lower() == 'self':
                self.parsed['url'] = clean(link.get('href', ''))
        else:
            self.parsed['rir'] = ''
        if self.parsed['url']:
            try:
                self.parsed['rir'] = self.bootstrap.get_rir_name_by_endpoint_url(
                    self.parsed['url'])
            except BootstrapError:
                pass

    def extract_entities(self, entities=None):
        self.parsed.setdefault('entities', {})

        if not entities:
            entities = self.raw_data.get('entities', [])

        for entity in entities:
            handle = clean(entity.get('handle', '')).upper()
            url = ''
            for link in entity.get('links', []):
                if link.get('rel', '').strip().lower() == 'self':
                    url = clean(link.get('href', ''))
            if not url:
                url = entity.get('url', '')
            rir = ''
            if url:
                try:
                    rir = self.bootstrap.get_rir_name_by_endpoint_url(url)
                except BootstrapError:
                    pass
            entity_type = clean(entity.get('objectClassName', ''))
            whois_server = clean(entity.get('port43', ''))
            # Most common use cases care about the name and email address
            vcard = self.parse_vcard_array(entity.get('vcardArray', []))
            for role in entity.get('roles', []):
                parsed_entity = {}
                if handle:
                    parsed_entity['handle'] = handle
                if url:
                    parsed_entity['url'] = url
                if entity_type:
                    parsed_entity['type'] = entity_type
                if whois_server:
                    parsed_entity['whois_server'] = whois_server
                if rir:
                    parsed_entity['rir'] = rir
                if vcard:
                    parsed_entity.update(vcard)
                if parsed_entity:
                    self.parsed['entities'].setdefault(role, [])
                    # ignore duplicate entity per role
                    if parsed_entity not in self.parsed['entities'].get(role, []):
                        self.parsed['entities'][role].append(parsed_entity)
            if entity.get('entities'):
                self.extract_entities(entities=entity.get('entities'))


class ParseAutnum(Parser):
    """
        Additional data extractors for autnum objects.
    """

    def parse(self):
        response_type = self.parsed['type']
        if response_type != 'autnum':
            raise ParseError(f'Expected response type of "autnum", got reply '
                             f'data of type "{response_type}" instead')
        self.extract_asn_range()
        return self.parsed

    def extract_asn_range(self):
        self.parsed['asn_range'] = None
        start_asn_range = self.raw_data.get('startAutnum', 0)
        end_asn_range = self.raw_data.get('endAutnum', 0)
        if start_asn_range > 0 and end_asn_range > 0:
            self.parsed['asn_range'] = [start_asn_range, end_asn_range]


class ParseDomain(Parser):
    """
        Additional data extractors for domain objects.
    """

    def parse(self):
        response_type = self.parsed['type']
        if response_type != 'domain':
            raise ParseError(f'Expected response type of "domain", got reply '
                             f'data of type "{response_type}" instead')
        self.extract_domain_name()
        self.extract_domain_nameservers()
        self.extract_domain_status()
        self.extract_domain_dnssec()
        return self.parsed

    def extract_domain_name(self):
        self.parsed['name'] = clean(self.raw_data.get('ldhName', ''))

    def extract_domain_nameservers(self):
        self.parsed['nameservers'] = []
        for nameserver in self.raw_data.get('nameservers', []):
            if nameserver.get('objectClassName', '') == 'nameserver':
                nameserver = nameserver.get('ldhName', '')
                if nameserver:
                    self.parsed['nameservers'].append(clean(nameserver))

    def extract_domain_status(self):
        self.parsed['status'] = []
        for status in self.raw_data.get('status', []):
            self.parsed['status'].append(clean(status))

    def extract_domain_dnssec(self):
        """
            SecureDNS.delegationSigned boolean indicates active dnssec.
        """
        self.parsed['dnssec'] = False
        if self.raw_data.get('secureDNS', {}).get('delegationSigned', None):
            self.parsed['dnssec'] = True


class ParseIPNetwork(Parser):
    """
        Additional data extractors for ip network objects.
    """

    def parse(self):
        response_type = self.parsed['type']
        if response_type != 'ip network':
            raise ParseError(f'Expected response type of "ip network", got reply '
                             f'data of type "{response_type}" instead')
        self.extract_country()
        self.extract_ip_version()
        self.extract_assignment_type()
        self.extract_network()
        return self.parsed

    def extract_country(self):
        self.parsed['country'] = clean(self.raw_data.get('country', ''))

    def extract_ip_version(self):
        self.parsed['ip_version'] = None
        ip_version = clean(self.raw_data.get('ipVersion', ''))
        if ip_version == 'v4':
            self.parsed['ip_version'] = 4
        elif ip_version == 'v6':
            self.parsed['ip_version'] = 6

    def extract_assignment_type(self):
        self.parsed['assignment_type'] = clean(self.raw_data.get('type', '')).lower()

    def extract_network(self):
        self.parsed['network'] = None
        cidr = self.raw_data.get('cidr0_cidrs', None)
        if isinstance(cidr, list):
            try:
                query_ip = ip_address(self.query.strip())
            except (ValueError, TypeError):
                query_ip = False
            for cidr_parts in cidr:
                length = cidr_parts.get('length', '')
                v4prefix = cidr_parts.get('v4prefix', '')
                v6prefix = cidr_parts.get('v6prefix', '')
                if length:
                    if v4prefix:
                        try:
                            parsed_network = IPv4Network(f'{v4prefix}/{length}')
                            if query_ip and query_ip.version == parsed_network.version:
                                if query_ip in parsed_network:
                                    self.parsed['network'] = parsed_network
                                    break
                            else:
                                self.parsed['network'] = parsed_network
                                break
                        except (TypeError, ValueError):
                            return
                    elif v6prefix:
                        try:
                            parsed_network = IPv6Network(f'{v6prefix}/{length}')
                            if query_ip and query_ip.version == parsed_network.version:
                                if query_ip in parsed_network:
                                    self.parsed['network'] = parsed_network
                                    break
                            else:
                                self.parsed['network'] = parsed_network
                                break
                        except (TypeError, ValueError):
                            return
        else:
            start_address = self.raw_data.get('startAddress', None)
            end_address = self.raw_data.get('endAddress', None)
            if start_address and end_address:
                try:
                    networks = summarize_address_range(ip_address(start_address),
                                                       ip_address(end_address))
                    self.parsed['network'] = list(networks)[0]
                except IndexError:
                    return


class ParseEntity(Parser):
    """
        Additional data extractors for entity objects.
    """

    def parse(self):
        response_type = self.parsed['type']
        if response_type != 'entity':
            raise ParseError(f'Expected response type of "entity", got reply '
                             f'data of type "{response_type}" instead')
        self.extract_root_vcard()
        return self.parsed

    def extract_root_vcard(self):
        root_vcard = self.raw_data.get('vcardArray', [])
        if root_vcard:
            parsed = self.parse_vcard_array(root_vcard)
            if parsed:
                self.parsed.update(parsed)


# These map the objectClassName values returned in RDAP responses
parser_map = {
    'autnum': ParseAutnum,
    'domain': ParseDomain,
    'ip network': ParseIPNetwork,
    'entity': ParseEntity,
}


def parse(bootstrap, data_type, query, raw_data):
    # Find a parser for the response type, falling back to the request / data type
    response_type = raw_data.get('objectClassName', data_type)
    parser_class = parser_map.get(response_type, None)
    if not parser_class:
        raise ParseError(f'No parser for response_type: {response_type}')
    log.debug(f'Parsing request type {data_type} {getsizeof(raw_data)} byte dict '
              f'with parser: {response_type} / {parser_class}')
    p = parser_class(bootstrap, raw_data, query,
                     using_overrides=bootstrap.is_using_overrides())
    return p.parse()



import requests
from .bootstrap import Bootstrap
from .query import QueryBuilder, Query
from .parser import parse
from .logger import get_logger
from .utils import create_session, get_session
from .version import version


# Private methods


_bootstrap = Bootstrap()
_query_builder = QueryBuilder(_bootstrap)


# Expose class methods as the public API


is_bootstrapped = _bootstrap.is_bootstrapped
clear_bootstrapping = _bootstrap.clear_bootstrapping
bootstrap = _bootstrap.bootstrap
save_bootstrap_data = _bootstrap.save_bootstrap_data
load_bootstrap_data = _bootstrap.load_bootstrap_data
bootstrap_is_older_than = _bootstrap.bootstrap_is_older_than
build_query = _query_builder.build


# Query helpers


def asn(as_number, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl=allow_insecure_ssl)
    method, url, exact_match = build_query(
        query_type='asn', query_value=as_number, rir=rir)
    q = Query(session, method, url)
    response = q.request()
    return response if raw else parse(_bootstrap, 'autnum', as_number, response)


def domain(domain_name, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl=allow_insecure_ssl)
    method, url, exact_match = build_query(
        query_type='domain', query_value=domain_name)
    q = Query(session, method, url)
    response = q.request()
    return response if raw else parse(_bootstrap, 'domain', domain_name, response)


def ip(ip_address_or_network, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl=allow_insecure_ssl)
    method, url, exact_match = build_query(
        query_type='ip', query_value=ip_address_or_network, rir=rir)
    q = Query(session, method, url)
    response = q.request()
    return response if raw else parse(_bootstrap, 'ip', ip_address_or_network, response)


def entity(entity_handle, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl=allow_insecure_ssl)
    method, url, exact_match = build_query(
        query_type='entity', query_value=entity_handle, rir=rir)
    q = Query(session, method, url)
    response = q.request()
    return response if raw else parse(_bootstrap, 'entity', entity_handle, response)



import requests
import unittest
import json
from pathlib import Path
from ipaddress import IPv4Network, IPv4Address, IPv6Network, IPv6Address
import whoisit


BASE_DIR = Path(__file__).resolve().parent


class QueryTestCase(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()
        with open(BASE_DIR / 'data_bootstrap.json', 'rt') as f:
            whoisit.load_bootstrap_data(f.read())

    def test_adding_query_url_params(self):
        url = 'https://example.com/'
        s = requests.Session()
        q = whoisit.query.Query(s, 'GET', url)
        params = {'test': 'test'}
        expected = 'https://example.com/?test=test'
        self.assertEqual(q.add_url_params(url, params), expected)
        params = {'test': 123}
        expected = 'https://example.com/?test=123'
        self.assertEqual(q.add_url_params(url, params), expected)

    def test_build(self):

        # Invalid build query requests
        with self.assertRaises(whoisit.errors.QueryError):
            whoisit.build_query(query_type='test', query_value=1)
        with self.assertRaises(whoisit.errors.QueryError):
            whoisit.build_query(query_type=b'test', query_value=1)
        with self.assertRaises(whoisit.errors.QueryError):
            whoisit.build_query(query_type={}, query_value=1)
        with self.assertRaises(whoisit.errors.QueryError):
            whoisit.build_query(query_type=None, query_value=1)
        with self.assertRaises(whoisit.errors.QueryError):
            whoisit.build_query()

        # ASN requests
        method, url, exact_match = whoisit.build_query(query_type='asn',
                                                       query_value=123)
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.arin.net/registry/autnum/123')
        method, url, exact_match = whoisit.build_query(query_type='asn',
                                                       query_value=61952)
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.db.ripe.net/autnum/61952')

        # Domain requests
        method, url, exact_match = whoisit.build_query(query_type='domain',
                                                       query_value='test.com')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.verisign.com/com/v1/domain/test.com')
        method, url, exact_match = whoisit.build_query(query_type='domain',
                                                       query_value='test.no')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.norid.no/domain/test.no')
        # TLD with no RDAP service should raise a UnsupportedError
        with self.assertRaises(whoisit.errors.UnsupportedError):
            whoisit.build_query(query_type='domain', query_value='test.test')

        # IP requests
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value='1.1.1.1')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.apnic.net/ip/1.1.1.1')
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value=IPv4Address('1.1.1.1'))
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.apnic.net/ip/1.1.1.1')
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value=IPv4Network('1.1.1.0/24'))
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.apnic.net/ip/1.1.1.0/24')
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value='2606:4700:4700::1111')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.arin.net/registry/ip/2606:4700:4700::1111')
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value=IPv6Address('2606:4700:4700::1111'))
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.arin.net/registry/ip/2606:4700:4700::1111')
        method, url, exact_match = whoisit.build_query(
            query_type='ip', query_value=IPv6Network('2606:4700::/32'))
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.arin.net/registry/ip/2606:4700::/32')

    def test_building_override_request(self):
        # Test that overriding the endpoint works, 1.1.1.1 is not allocated to afrinic
        method, url, exact_match = whoisit.build_query(query_type='ip',
                                                       query_value='1.1.1.1',
                                                       rir='afrinic')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.afrinic.net/rdap/ip/1.1.1.1')
        # ... or ripe
        method, url, exact_match = whoisit.build_query(query_type='ip',
                                                       query_value='1.1.1.1',
                                                       rir='ripe')
        self.assertEqual(method, 'GET')
        self.assertEqual(url, 'https://rdap.db.ripe.net/ip/1.1.1.1')



import unittest
import json
from pathlib import Path
from datetime import datetime
from ipaddress import IPv4Network, IPv6Network
from dateutil.tz import tzoffset, tzutc
import whoisit


BASE_DIR = Path(__file__).resolve().parent


class ParserTestCase(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()
        with open(BASE_DIR / 'data_bootstrap.json', 'rt') as f:
            whoisit.load_bootstrap_data(f.read())

    def test_parser_interface(self):
        with self.assertRaises(whoisit.errors.ParseError):
            whoisit.parser.parse(whoisit._bootstrap, 'invalid', '', {})
        fake_response = {'handle': 'TEST', 'name': 'test'}
        fake_response['objectClassName'] = 'autnum'
        whoisit.parser.parse(whoisit._bootstrap, 'autnum', '', fake_response)
        fake_response['objectClassName'] = 'domain'
        whoisit.parser.parse(whoisit._bootstrap, 'domain', '', fake_response)
        fake_response['objectClassName'] = 'ip network'
        whoisit.parser.parse(whoisit._bootstrap, 'ip', '', fake_response)
        fake_response['objectClassName'] = 'entity'
        whoisit.parser.parse(whoisit._bootstrap, 'entity', '', fake_response)

    def test_autnum_response_parser(self):
        with open(BASE_DIR / 'data_rdap_response_asn.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'autnum', '13335', test_data)
        self.assertEqual(parsed['type'], 'autnum')
        self.assertEqual(parsed['handle'], 'AS13335')
        self.assertEqual(parsed['name'], 'CLOUDFLARENET')
        self.assertEqual(parsed['url'], 'https://rdap.arin.net/registry/autnum/13335')
        self.assertEqual(parsed['handle'], 'AS13335')
        self.assertEqual(parsed['rir'], 'arin')
        self.assertEqual(parsed['asn_range'], [13335, 13335])
        self.assertEqual(parsed['whois_server'], 'whois.arin.net')
        self.assertEqual(parsed['copyright_notice'], 'Copyright 1997-2021, American Registry for Internet Numbers, Ltd.')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.arin.net/resources/registry/whois/tou/')
        self.assertEqual(parsed['last_changed_date'], datetime(2017, 2, 17, 18, 4, 32, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['registration_date'], datetime(2010, 7, 14, 18, 35, 57, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(parsed['entities']['registrant'],
            [
                {
                    'handle': 'CLOUD14',
                    'url': 'https://rdap.arin.net/registry/entity/CLOUD14',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'Cloudflare, Inc.',
                    'rir': 'arin'
                }
            ]
        )
        self.assertEqual(parsed['entities']['abuse'],
            [
                {
                    'handle': 'ABUSE2916-ARIN',
                    'url': 'https://rdap.arin.net/registry/entity/ABUSE2916-ARIN',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'Abuse',
                    'email': 'abuse@cloudflare.com',
                    'tel': '+1-650-319-8930',
                    'rir': 'arin'
                }
            ]
        )
        self.assertEqual(parsed['entities']['noc'],
            [
                {
                    'handle': 'NOC11962-ARIN',
                    'url': 'https://rdap.arin.net/registry/entity/NOC11962-ARIN',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'NOC',
                    'email': 'noc@cloudflare.com',
                    'tel': '+1-650-319-8930',
                    'rir': 'arin'
                }
            ]
        )
        self.assertEqual(parsed['entities']['technical'],
            [
                {
                    'handle': 'ADMIN2521-ARIN',
                    'url': 'https://rdap.arin.net/registry/entity/ADMIN2521-ARIN',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'Admin',
                    'email': 'rir@cloudflare.com',
                    'tel': '+1-650-319-8930',
                    'rir': 'arin'
                }
            ]
        )

    def test_domain_response_parser(self):

        # google.com
        with open(BASE_DIR / 'data_rdap_response_domain1.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'domain', 'GOOGLE.COM', test_data)
        self.assertEqual(parsed['type'], 'domain')
        self.assertEqual(parsed['name'], 'GOOGLE.COM')
        self.assertEqual(parsed['handle'], '2138514_DOMAIN_COM-VRSN')
        self.assertEqual(parsed['rir'], '')
        self.assertEqual(parsed['last_changed_date'], None)
        self.assertEqual(parsed['registration_date'], datetime(1997, 9, 15, 4, 0, tzinfo=tzutc()))
        self.assertEqual(parsed['expiration_date'], datetime(2028, 9, 14, 4, 0, tzinfo=tzutc()))
        self.assertEqual(parsed['url'], 'https://rdap.verisign.com/com/v1/domain/GOOGLE.COM')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.verisign.com/domain-names/registration-data-access-protocol/terms-service/index.xhtml')
        self.assertEqual(parsed['whois_server'], '')
        self.assertEqual(parsed['copyright_notice'], '')
        self.assertEqual(parsed['dnssec'], True)
        self.assertEqual(parsed['nameservers'], [
            'NS1.GOOGLE.COM',
            'NS2.GOOGLE.COM',
            'NS3.GOOGLE.COM',
            'NS4.GOOGLE.COM'
        ])
        self.assertEqual(parsed['status'], [
            'client delete prohibited',
            'client transfer prohibited',
            'client update prohibited',
            'server delete prohibited',
            'server transfer prohibited',
            'server update prohibited'
        ])
        self.assertEqual(parsed['entities']['registrar'],
            [
                {
                    'handle': '292',
                    'type': 'entity',
                    'name': 'MarkMonitor Inc.'
                }
            ]
        )

        # norway.no
        with open(BASE_DIR / 'data_rdap_response_domain2.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'domain', 'norway.no', test_data)
        self.assertEqual(parsed['type'], 'domain')
        self.assertEqual(parsed['name'], 'norway.no')
        self.assertEqual(parsed['handle'], 'NOR34044D-NORID')
        self.assertEqual(parsed['rir'], '')
        self.assertEqual(parsed['last_changed_date'], datetime(2021, 1, 14, 5, 16, 59, tzinfo=tzutc()))
        self.assertEqual(parsed['registration_date'], datetime(2017, 1, 24, 12, 9, 23, tzinfo=tzutc()))
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.norid.no/domain/norway.no')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.norid.no/en/domeneoppslag/vilkar')
        self.assertEqual(parsed['whois_server'], '')
        self.assertEqual(parsed['copyright_notice'], '')
        self.assertEqual(parsed['nameservers'], [
            'ns1-09.azure-dns.com',
            'ns2-09.azure-dns.net',
            'ns3-09.azure-dns.org',
            'ns4-09.azure-dns.info'
        ])
        self.assertEqual(parsed['status'], [])
        self.assertEqual(parsed['entities']['registrar'],
            [
                {
                    'handle': 'REG42-NORID',
                    'url': 'https://rdap.norid.no/entity/reg42-NORID',
                    'type': 'entity',
                    'name': 'Domeneshop AS',
                    'email': 'kundeservice@domeneshop.no'
                }
            ]
        )
        self.assertEqual(parsed['entities']['technical'],
            [
                {
                    'handle': 'DH21326R-NORID',
                    'url': 'https://rdap.norid.no/entity/DH21326R-NORID',
                    'type': 'entity',
                    'name': 'Domeneshop Hostmaster',
                    'email': 'hostmaster@domeneshop.no'
                }
            ]
        )

        # themarquetry.com
        with open(BASE_DIR / 'data_rdap_response_domain3.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'domain', 'THEMARQUETRY.COM', test_data)
        self.assertEqual(parsed['type'], 'domain')
        self.assertEqual(parsed['name'], 'THEMARQUETRY.COM')
        self.assertEqual(parsed['handle'], '2598322308_DOMAIN_COM-VRSN')
        self.assertEqual(parsed['rir'], '')
        self.assertEqual(parsed['nameservers'], [
            'NS1.DNS-PARKING.COM',
            'NS2.DNS-PARKING.COM'
        ])
        self.assertEqual(parsed['status'], ['client transfer prohibited'])
        self.assertEqual(parsed['entities']['registrar'],
            [
                {
                    'handle': '1647',
                    'type': 'entity',
                    'name': 'Hosting Concepts B.V. d/b/a Registrar.eu',
                }
            ]
        )
        self.assertEqual(parsed['entities']['abuse'],
            [
                {
                    'type': 'entity',
                    'name': '',
                    'email': 'abuse@registrar.eu'
                }
            ]
        )

    def test_ip_response_parser(self):

        # ipv4 address
        with open(BASE_DIR / 'data_rdap_response_ip_v4.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '1.1.1.1', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'APNIC-LABS')
        self.assertEqual(parsed['handle'], '1.1.1.0 - 1.1.1.255')
        self.assertEqual(parsed['rir'], 'apnic')
        self.assertEqual(parsed['last_changed_date'], datetime(2020, 7, 15, 13, 10, 57, tzinfo=tzutc()))
        self.assertEqual(parsed['registration_date'], None)
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.apnic.net/ip/1.1.1.0/24')
        self.assertEqual(parsed['terms_of_service_url'], 'http://www.apnic.net/db/dbcopyright.html')
        self.assertEqual(parsed['whois_server'], 'whois.apnic.net')
        self.assertEqual(parsed['copyright_notice'], '')
        self.assertEqual(parsed['description'], [
            'APNIC and Cloudflare DNS Resolver project',
            'Routed globally by AS13335/Cloudflare',
            'Research prefix for APNIC Labs'
        ])
        self.assertEqual(parsed['entities']['administrative'],
            [
                {
                    'handle': 'AR302-AP',
                    'url': 'https://rdap.apnic.net/entity/AR302-AP',
                    'type': 'entity',
                    'name': 'APNIC RESEARCH',
                    'email': 'research@apnic.net',
                    'tel': '+61-7-3858-3199',
                    'rir': 'apnic'
                }
            ]
        )

        # ipv4 address 2
        with open(BASE_DIR / 'data_rdap_response_ip_v4_2.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '130.59.31.80', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'SWITCH-LAN')
        self.assertEqual(parsed['rir'], 'ripe')
        self.assertEqual(parsed['url'], 'https://rdap.db.ripe.net/ip/130.59.31.80')
        self.assertEqual(parsed['whois_server'], 'whois.ripe.net')
        self.assertEqual(parsed['network'], IPv4Network('130.59.0.0/16'))

        # ipv4 address 3
        with open(BASE_DIR / 'data_rdap_response_ip_v4_3.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '13.104.0.0', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'MSFT')
        self.assertEqual(parsed['rir'], 'arin')
        self.assertEqual(parsed['url'], 'https://rdap.arin.net/registry/ip/13.64.0.0')
        self.assertEqual(parsed['whois_server'], 'whois.arin.net')
        self.assertEqual(parsed['network'], IPv4Network('13.104.0.0/14'))

        # ipv4 network
        with open(BASE_DIR / 'data_rdap_response_cidr_v4.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '1.1.1.0/24', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'APNIC-LABS')
        self.assertEqual(parsed['handle'], '1.1.1.0 - 1.1.1.255')
        self.assertEqual(parsed['rir'], 'apnic')
        self.assertEqual(parsed['last_changed_date'], datetime(2020, 7, 15, 13, 10, 57, tzinfo=tzutc()))
        self.assertEqual(parsed['registration_date'], None)
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.apnic.net/ip/1.1.1.0/24')
        self.assertEqual(parsed['terms_of_service_url'], 'http://www.apnic.net/db/dbcopyright.html')
        self.assertEqual(parsed['whois_server'], 'whois.apnic.net')
        self.assertEqual(parsed['copyright_notice'], '')
        self.assertEqual(parsed['assignment_type'], 'assigned portable')
        self.assertEqual(parsed['country'], 'AU')
        self.assertEqual(parsed['ip_version'], 4)
        self.assertEqual(parsed['network'], IPv4Network('1.1.1.0/24'))
        self.assertEqual(parsed['description'], [
            'APNIC and Cloudflare DNS Resolver project',
            'Routed globally by AS13335/Cloudflare',
            'Research prefix for APNIC Labs'
        ])
        self.assertEqual(parsed['entities']['administrative'],
            [
                {
                    'handle': 'AR302-AP',
                    'url': 'https://rdap.apnic.net/entity/AR302-AP',
                    'type': 'entity',
                    'name': 'APNIC RESEARCH',
                    'email': 'research@apnic.net',
                    'tel': '+61-7-3858-3199',
                    'rir': 'apnic'
                }
            ]
        )

        # ipv6 address
        with open(BASE_DIR / 'data_rdap_response_ip_v6.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '2001:4860::', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'GOOGLE-IPV6')
        self.assertEqual(parsed['handle'], 'NET6-2001-4860-1')
        self.assertEqual(parsed['parent_handle'], 'NET6-2001-4800-0')
        self.assertEqual(parsed['rir'], 'arin')
        self.assertEqual(parsed['last_changed_date'], datetime(2012, 2, 24, 9, 44, 34, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['registration_date'], datetime(2005, 3, 14, 11, 31, 8, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.arin.net/registry/ip/2001:4860::')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.arin.net/resources/registry/whois/tou/')
        self.assertEqual(parsed['whois_server'], 'whois.arin.net')
        self.assertEqual(parsed['copyright_notice'], 'Copyright 1997-2021, American Registry for Internet Numbers, Ltd.')
        self.assertEqual(parsed['assignment_type'], 'direct allocation')
        self.assertEqual(parsed['ip_version'], 6)
        self.assertEqual(parsed['network'], IPv6Network('2001:4860::/32'))
        self.assertEqual(parsed['description'], [])
        self.assertEqual(parsed['entities']['registrant'],
            [
                {
                    'handle': 'GOGL',
                    'name': 'Google LLC',
                    'rir': 'arin',
                    'type': 'entity',
                    'url': 'https://rdap.arin.net/registry/entity/GOGL',
                    'whois_server': 'whois.arin.net'
                }
            ]
        )
        self.assertEqual(parsed['entities']['noc'],
            [
                {
                    'email': 'arin-contact@google.com',
                    'handle': 'ZG39-ARIN',
                    'name': 'Google LLC',
                    'rir': 'arin',
                    'type': 'entity',
                    'url': 'https://rdap.arin.net/registry/entity/ZG39-ARIN',
                    'tel': '+1-650-253-0000',
                    'whois_server': 'whois.arin.net'
                }
            ]
        )
        self.assertEqual(parsed['entities']['technical'],
            [
                {
                    'email': 'arin-contact@google.com',
                    'handle': 'ZG39-ARIN',
                    'name': 'Google LLC',
                    'rir': 'arin',
                    'type': 'entity',
                    'url': 'https://rdap.arin.net/registry/entity/ZG39-ARIN',
                    'tel': '+1-650-253-0000',
                    'whois_server': 'whois.arin.net'
                }
            ]
        )
        self.assertEqual(parsed['entities']['abuse'],
            [
                {
                    'handle': 'ABUSE5250-ARIN',
                    'url': 'https://rdap.arin.net/registry/entity/ABUSE5250-ARIN',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'Abuse',
                    'email': 'network-abuse@google.com',
                    'tel': '+1-650-253-0000',
                    'rir': 'arin'
                },
                {
                    'handle': 'ZG39-ARIN',
                    'url': 'https://rdap.arin.net/registry/entity/ZG39-ARIN',
                    'type': 'entity',
                    'whois_server': 'whois.arin.net',
                    'name': 'Google LLC',
                    'email': 'arin-contact@google.com',
                    'tel': '+1-650-253-0000',
                    'rir': 'arin'
                }
            ]
        )
        # ipv6 network
        with open(BASE_DIR / 'data_rdap_response_cidr_v6.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', '2001:4860::', test_data)
        self.assertEqual(parsed['type'], 'ip network')
        self.assertEqual(parsed['name'], 'GOOGLE-IPV6')
        self.assertEqual(parsed['handle'], 'NET6-2001-4860-1')
        self.assertEqual(parsed['parent_handle'], 'NET6-2001-4800-0')
        self.assertEqual(parsed['rir'], 'arin')
        self.assertEqual(parsed['last_changed_date'], datetime(2012, 2, 24, 9, 44, 34, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['registration_date'], datetime(2005, 3, 14, 11, 31, 8, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.arin.net/registry/ip/2001:4860::')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.arin.net/resources/registry/whois/tou/')
        self.assertEqual(parsed['whois_server'], 'whois.arin.net')
        self.assertEqual(parsed['copyright_notice'], 'Copyright 1997-2021, American Registry for Internet Numbers, Ltd.')
        self.assertEqual(parsed['assignment_type'], 'direct allocation')
        self.assertEqual(parsed['ip_version'], 6)
        self.assertEqual(parsed['network'], IPv6Network('2001:4860::/32'))
        self.assertEqual(parsed['description'], [])
        self.assertEqual(parsed['entities']['registrant'],
            [
                {
                    'handle': 'GOGL',
                    'name': 'Google LLC',
                    'rir': 'arin',
                    'type': 'entity',
                    'url': 'https://rdap.arin.net/registry/entity/GOGL',
                    'whois_server': 'whois.arin.net'
                }
            ]
        )

    def test_entity_response_parser(self):
        with open(BASE_DIR / 'data_rdap_response_entity.json') as f:
            test_data = json.loads(f.read())
        parsed = whoisit.parser.parse(whoisit._bootstrap, 'ip', 'GOVI', test_data)
        self.assertEqual(parsed['type'], 'entity')
        self.assertEqual(parsed['name'], 'Govital Internet Inc.')
        self.assertEqual(parsed['handle'], 'GOVI')
        self.assertEqual(parsed['parent_handle'], '')
        self.assertEqual(parsed['rir'], 'arin')
        self.assertEqual(parsed['last_changed_date'], datetime(2017, 1, 28, 8, 32, 29, tzinfo=tzoffset(None, -18000)))
        self.assertEqual(parsed['registration_date'], datetime(2001, 5, 8, 0, 0, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(parsed['expiration_date'], None)
        self.assertEqual(parsed['url'], 'https://rdap.arin.net/registry/entity/GOVI')
        self.assertEqual(parsed['terms_of_service_url'], 'https://www.arin.net/resources/registry/whois/tou/')
        self.assertEqual(parsed['whois_server'], 'whois.arin.net')
        self.assertEqual(parsed['copyright_notice'], 'Copyright 1997-2021, American Registry for Internet Numbers, Ltd.')
        self.assertEqual(parsed['description'], ['http://www.govital.net\\r', 'Standard NOC hours are 10am to 6pm EST M-F'])
        self.assertEqual(parsed['entities']['technical'],
            [
                {
                    'email': 'support@govital.net',
                    'handle': 'GTS7-ARIN',
                    'name': 'Govital Technical Support',
                    'rir': 'arin',
                    'type': 'entity',
                    'url': 'https://rdap.arin.net/registry/entity/GTS7-ARIN',
                    'tel': '+1-519-254-5115',
                    'whois_server': 'whois.arin.net'
                }
            ]
        )



import os
import sys
from setuptools import setup, find_packages


version = '2.7.3'


with open('README.md', 'rt') as f:
    long_description = f.read()


with open('requirements.txt', 'rt') as f:
    requirements = tuple(f.read().split())


setup(
    name = 'whoisit',
    version = version,
    url = 'https://github.com/meeb/whoisit',
    author = 'https://github.com/meeb',
    author_email = 'meeb@meeb.org',
    description = 'A Python client to RDAP WHOIS-like services for internet resources.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license = 'BSD',
    include_package_data = True,
    install_requires = requirements,
    packages = find_packages(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords = ['whoisit', 'whois', 'rdap', 'ip', 'network', 'cidr', 'prefix', 'domain',
                'asn', 'autnum', 'tld', 'entity', 'handle', 'arin', 'afrinic', 'apnic',
                'ripe', 'lacnic']
)



import random
from urllib.parse import (urlsplit, urlunsplit, parse_qs, urljoin, urlencode, quote,
                          unquote)
from ipaddress import (ip_address, ip_network, IPv4Address, IPv4Network, IPv6Address,
                       IPv6Network)
from .utils import http_request, contains_only_chars
from .logger import get_logger
from .errors import (QueryError, ResourceDoesNotExist, UnsupportedError,
                     RateLimitedError, RemoteServerError)


log = get_logger('query')


class QueryBuilder:

    QUERY_TYPES_MAP = {
        'asn': 'autnum',
        'as': 'autnum',
        'autnum': 'autnum',
        'domain': 'domain',
        'tld': 'domain',
        'dns': 'domain',
        'ipv4': 'ip',
        'ipv6': 'ip',
        'ip': 'ip',
        'prefix': 'ip',
        'cidr': 'ip',
        'network': 'ip',
        'object': 'entity',
        'entity': 'entity',
    }

    def __init__(self, bootstrap):
        # These map to the supported RDAP types, not the bootstrap data types
        self.query_endpoints_fetchers = {
            'autnum': self.get_autnum_endpoint,
            'domain': self.get_domain_endpoint,
            'ip': self.get_ip_endpoint,
            'entity': self.get_entity_endpoint,
        }
        self.bootstrap = bootstrap

    def build(self, query_type=None, query_value=None, rir=None):
        if not self.bootstrap.is_bootstrapped():
            raise QueryError(f'You need to load bootstrap data before making '
                             f'any queries')
        query_type = str(query_type).strip().lower()
        what = self.QUERY_TYPES_MAP.get(query_type)
        if not what:
            raise QueryError(f'Unknown query_type: {query_type}')
        if not query_value:
            raise QueryError(f'query_value must be set')
        if rir:
            url, exact_match = self.get_override_endpoint(rir, what, query_value)
        else:
            fetcher = self.query_endpoints_fetchers[what]
            url, exact_match = fetcher(query_value)
        method = 'GET'
        match_str = ' (exact match)' if exact_match else ''
        log.debug(f'{what} query for {query_value} built as {method} {url}{match_str}')
        return method, url, exact_match

    def construct_url(self, base_url, what, value):
        if not base_url.endswith('/'):
            base_url += '/'
        resource = urljoin(base_url, str(what))
        if not resource.endswith('/'):
            resource += '/'
        quoted_value = quote(str(value))
        return unquote(urljoin(resource, quoted_value))

    def get_autnum_endpoint(self, value):
        if isinstance(value, str):
            value = value.strip().upper()
            if value.startswith('AS'):
                value = value[2:]
        try:
            value = int(value)
        except (TypeError, ValueError) as e:
            raise QueryError(f'Failed to cast AS number to integer: {e}') from e
        endpoints, exact_match = self.bootstrap.get_asn_endpoints(value)
        endpoint = random.choice(endpoints)
        if not exact_match:
            log.debug(f'Failed to match ASN: {value} to an RDAP service, '
                      f'defaulting to: {endpoint}')
        return self.construct_url(endpoint, 'autnum', value), exact_match

    def get_domain_endpoint(self, value):
        value = str(value).strip()
        parts = value.split('.')
        if len(parts) < 2:
            raise QueryError(f'Failed to extract TLD from domain "{value}"')
        domain = '.'.join(parts[:-1])
        try:
            # First check for SLDs
            tld = '.'.join(parts[-2:])
            endpoints, exact_match = self.bootstrap.get_dns_endpoints(tld)
        except UnsupportedError:
            # If no SLD match, try the TLD
            tld = parts[-1]
            endpoints, exact_match = self.bootstrap.get_dns_endpoints(tld)
        endpoint = random.choice(endpoints)
        if not exact_match:
            log.debug(f'Failed to match domain: {value} to an RDAP service, '
                      f'defaulting to: {endpoint}')
        return self.construct_url(endpoint, 'domain', value), exact_match

    def get_ip_endpoint(self, value):
        if not isinstance(value, (IPv4Address, IPv4Network, IPv6Address, IPv6Network)):
            value = str(value)
        if isinstance(value, str):
            try:
                value = ip_address(value)
            except (TypeError, ValueError):
                try:
                    value = ip_network(value)
                except (TypeError, ValueError):
                    raise QueryError(f'Unable to cast input as either an IP address '
                                     f'or IP network: {value}')
        if value.version == 4:
            endpoints, exact_match = self.bootstrap.get_ipv4_endpoints(value)
            endpoint = random.choice(endpoints)
            if not exact_match:
                log.debug(f'Failed to match IPv4: {value} to an RDAP service, '
                          f'defaulting to: {endpoint}')
        else:
            endpoints, exact_match = self.bootstrap.get_ipv6_endpoints(value)
            endpoint = random.choice(endpoints)
            if not exact_match:
                log.debug(f'Failed to match IPv6: {value} to an RDAP service, '
                          f'defaulting to: {endpoint}')
        return self.construct_url(endpoint, 'ip', str(value)), exact_match

    def get_entity_endpoint(self, value):
        value = str(value).strip().upper()
        # Entity names can only contain A-Z upper case, 0-9 and hyphens
        if not contains_only_chars(value):
            raise QueryError(f'Entity tags or handles can only container uppercase '
                             f'A-Z, 0-9 and hyphens, got: {value}')
        endpoints, exact_match = self.bootstrap.get_entity_endpoints(value)
        endpoint = random.choice(endpoints)
        return self.construct_url(endpoint, 'entity', value), exact_match

    def get_override_endpoint(self, rir_endpoint_name, query_name, value):
        '''
            The query should be built using a manually overriden RIR endpoint name,
            such as 'arin' or 'ripe' and not use the bootstrap data.
        '''
        if isinstance(value, str):
            value = value.strip()
        query_name = str(query_name).strip()
        endpoint = self.bootstrap.get_rir_endpoint(rir_endpoint_name)
        if not endpoint:
            rir_names = self.bootstrap.get_rir_endpoint_names()
            raise QueryError(f'Unknown RIR endpoint name, must be one of: {rir_names}')
        return self.construct_url(endpoint, query_name, value), True


class Query:
    """
        Make an HTTP request to an RDAP endpoint as a query. This is slightly more
        elaborate than a single function just to allow kwargs to be arbitrarily passed
        to both requests and the requested URL if required.
    """

    def __init__(self, session, method, url, **kwargs):
        self.session = session
        self.method = method.strip().upper()
        if kwargs:
            # kwargs are appended to the URL, such as test=123 becomes url?test=123
            self.url = self.add_url_params(url, kwargs)
        else:
            self.url = url

    def add_url_params(self, url, extra_params):
        parts = urlsplit(url)
        qs = {}
        for k, v in parse_qs(parts.query):
            qs[k] = v
        for k, v in extra_params.items():
            qs[k] = str(v)
        qs_str = urlencode(qs)
        return urlunsplit((parts.scheme, parts.netloc, parts.path, qs_str, ''))

    def request(self, *args, **kwargs):
        # args and kwargs here are passed directly to requests.request(...)
        response = http_request(self.session, url=self.url, method=self.method, *args, **kwargs)
        if response.status_code == 404:
            raise ResourceDoesNotExist(f'RDAP {self.method} request to {self.url} '
                                       f'returned a 404 error, the resource does '
                                       f'not exist')
        elif response.status_code == 429:
            raise RateLimitedError(f'RDAP {self.method} request to {self.url} '
                                   f'returned a 429 error, the resource has been '
                                   f'rate limited')
        elif response.status_code == 500:
            raise RemoteServerError(f'RDAP {self.method} request to {self.url} '
                                   f'returned a 500 error, the resource returned '
                                   f'a remote server error')
        elif response.status_code != 200:
            raise QueryError(f'RDAP {self.method} request to {self.url} returned a '
                             f'non-200 status code of {response.status_code}')
        try:
            return response.json()
        except (TypeError, ValueError) as e:
            raise QueryError(f'Failed to parse RDAP Query response as JSON: {e}') from e
